"""
This script reads the MUSDB lyrics annotation files,
cuts the audio into snippets according to annotated lines,
and saves audio and text files accordingly.

It requires the python packages musdb (https://github.com/sigsep/sigsep-mus-db)
and soundfile (https://pypi.org/project/SoundFile/).

The script has been tested using Python 3.6 on macOS.
"""
import os
import yaml

import musdb
import soundfile as sf

# ------------------------------------------------------------------------------------------------------------------
# THESE VARIABLES NEED TO BE ADAPTED BY THE USER
path_to_musdb = '../Datasets/MUSDB18'  # folder where MUSDB18 mp4 files are stored
path_to_train_lyrics = '../Datasets/MUSDB_lyrics/train_lyrics'  # folder of lyrics for training set
path_to_test_lyrics = '../Datasets/MUSDB_lyrics/test_lyrics'  # folder of lyrics for test set

path_to_save_data = '../Datasets/MUSDB_lyrics/'  # folder to save wav files and text files

sources = ['mixture', 'accompaniment', 'vocals', 'bass', 'drums', 'other']  # select sources you want to save as wav
vocals_types = ['a', 'b', 'c', 'd', 'rejected']  # select vocals types you want to consider

# ------------------------------------------------------------------------------------------------------------------
# Determine validation set

# Same validation set as the original MUSDB validation set
# (mus_valid = musdb_corpus.load_mus_tracks(subsets="train", split='valid'))
# but tracks with non-English lyrics were replaced by another track from the same artist:
# Fergessen - Nos Palpitants replaced by Fergessen - The Wind
# Meaxic - Take A Step replaced by Meaxic - You Listen
#
# If no validation set is needed, provide an empty list (validation_tracks = [])
validation_tracks = ['Actions - One Minute Smile',
                     'Clara Berry And Wooldog - Waltz For My Victims',
                     'Johnny Lokke - Promises & Lies',
                     'Patrick Talbot - A Reason To Leave',
                     'Triviul - Angelsaint',
                     'Alexander Ross - Goodbye Bolero',
                     'Fergessen - The Wind',
                     'Leaf - Summerghost',
                     'Skelpolu - Human Mistakes',
                     'Young Griffo - Pennies',
                     'ANiMAL - Rockshow',
                     'James May - On The Line',
                     'Meaxic - You Listen',
                     'Traffic Experiment - Sirens']

# ------------------------------------------------------------------------------------------------------------------

# Ignore warning about unsafe loaders in pyYAML 5.1 (used in musdb)
# https://github.com/yaml/pyyaml/wiki/PyYAML-yaml.load(input)-Deprecation
yaml.warnings({'YAMLLoadWarning': False})

# ------------------------------------------------------------------------------------------------------------------
# Make folder structure
path = os.path.join(path_to_save_data, 'test', 'text')
if not os.path.isdir(path): os.makedirs(path, exist_ok=True)
for source in sources:
    for type in vocals_types:
        path = os.path.join(path_to_save_data, 'test', 'audio', source, type)
        if not os.path.isdir(path): os.makedirs(path, exist_ok=True)

path = os.path.join(path_to_save_data, 'train', 'text')
if not os.path.isdir(path): os.makedirs(path, exist_ok=True)
for source in sources:
    for type in vocals_types:
        path = os.path.join(path_to_save_data, 'train', 'audio', source, type)
        if not os.path.isdir(path): os.makedirs(path, exist_ok=True)

path = os.path.join(path_to_save_data, 'val', 'text')
if not os.path.isdir(path): os.makedirs(path, exist_ok=True)
for source in sources:
    for type in vocals_types:
        path = os.path.join(path_to_save_data, 'val', 'audio', source, type)
        if not os.path.isdir(path): os.makedirs(path, exist_ok=True)
# ------------------------------------------------------------------------------------------------------------------
# Load MUSDB
musdb_corpus = musdb.DB(path_to_musdb)
training_tracks = musdb_corpus.load_mus_tracks(subsets=['train'])
test_tracks = musdb_corpus.load_mus_tracks(subsets=['test'])

# -----------------------------------------------------------------------------------------------------------------
# Process MUSDB test partition and save test files
for track in test_tracks:

    track_name = track.name
    print(track_name)

    path_to_track_lyrics = os.path.join(path_to_test_lyrics, track_name + '.txt')

    # ignore files without lyrics annotations
    if not os.path.isfile(path_to_track_lyrics):
        print("No lyrics for", track, ". It has probably lyrics that are not in English. It was skipped")
        continue

    # read lyrics file
    lyrics_file = open(path_to_track_lyrics)
    lyrics_lines = lyrics_file.readlines()

    # go through lyrics lines and split audio as annotated
    for counter, line in enumerate(lyrics_lines):

        # handle rejected lines
        if line[0] == '*':
            if 'rejected' in vocals_types:
                annotations = line.split(' ', maxsplit=4)
                _, start, end, vocals_type, lyrics = annotations
            else:
                continue
        else:
            annotations = line.split(' ', maxsplit=3)
            start, end, vocals_type, lyrics = annotations
            if vocals_type not in vocals_types:
                continue

        start_m = int(start.split(':')[0])  # start time minutes
        start_s = int(start.split(':')[1])  # start time seconds
        start_time = start_m * 60 + start_s  # start time in seconds

        end_m = int(end.split(':')[0])  # end time minutes
        end_s = int(end.split(':')[1])  # end time seconds
        end_time = end_m * 60 + end_s  # end time in seconds

        length = end_time - start_time

        # set track duration and start in order ton load only the desired snippet
        track.chunk_duration = length
        track.chunk_start = start_time

        snippet_file_name = track_name + '_{}'.format(counter)
        lyrics = lyrics.replace('\n', '')
        partition = 'test'

        # write wav files
        for source in sources:
            if source == 'mixture': audio = track.audio
            else: audio = track.targets[source].audio
            sf.write(os.path.join(path_to_save_data, partition, 'audio', source,
                                  vocals_type, snippet_file_name + '.wav'), audio, samplerate=track.rate)

        # write text file
        path_to_save_text = os.path.join(path_to_save_data, partition, 'text', snippet_file_name + '.txt')
        with open(path_to_save_text, 'w') as txt_file:
            txt_file.write(lyrics)
            txt_file.close()

# -----------------------------------------------------------------------------------------------------------------
# Process MUSDB training partition and save training and validation files
for track in training_tracks:

    track_name = track.name
    print(track_name)

    path_to_track_lyrics = os.path.join(path_to_train_lyrics, track_name + '.txt')

    # ignore files without lyrics annotations
    if not os.path.isfile(path_to_track_lyrics):
        print("No lyrics for", track, ". It has probably lyrics that are not in English. It was skipped")
        continue

    # read lyrics file
    lyrics_file = open(path_to_track_lyrics)
    lyrics_lines = lyrics_file.readlines()

    # go through lyrics lines and split audio as annotated
    for counter, line in enumerate(lyrics_lines):

        # handle rejected lines
        if line[0] == '*':
            if 'rejected' in vocals_types:
                annotations = line.split(' ', maxsplit=4)
                _, start, end, vocals_type, lyrics = annotations
            else:
                continue
        else:
            annotations = line.split(' ', maxsplit=3)
            start, end, vocals_type, lyrics = annotations
            if vocals_type not in vocals_types:
                continue

        start_m = int(start.split(':')[0])  # start time minutes
        start_s = int(start.split(':')[1])  # start time seconds
        start_time = start_m * 60 + start_s  # start time in seconds

        end_m = int(end.split(':')[0])  # end time minutes
        end_s = int(end.split(':')[1])  # end time seconds
        end_time = end_m * 60 + end_s  # end time in seconds

        length = end_time - start_time

        # set track duration and start in order ton load only the desired snippet
        track.chunk_duration = length
        track.chunk_start = start_time

        snippet_file_name = track_name + '_{}'.format(counter)
        lyrics = lyrics.replace('\n', '')
        partition = 'val' if track_name in validation_tracks else 'train'

        # write wav files
        for source in sources:
            if source == 'mixture': audio = track.audio
            else: audio = track.targets[source].audio
            sf.write(os.path.join(path_to_save_data, partition, 'audio', source,
                                  vocals_type, snippet_file_name + '.wav'), audio, samplerate=track.rate)

        # write text file
        path_to_save_text = os.path.join(path_to_save_data, partition, 'text', snippet_file_name + '.txt')
        with open(path_to_save_text, 'w') as txt_file:
            txt_file.write(lyrics)
            txt_file.close()
